# Import classes from .pubsub that used to be in .units for backwards compatibility.
from .pubsub import ZMQPollerUnit as ZMQPollerUnit
from .pubsub import ZMQPollerState as ZMQPollerState
from .pubsub import ZMQPollerSettings as ZMQPollerSettings
from .pubsub import ZMQSenderUnit as ZMQSenderUnit
from .pubsub import ZMQSenderState as ZMQSenderState
from .pubsub import ZMQSenderSettings as ZMQSenderSettings
from .util import ZMQMessage as ZMQMessage



import json
import threading
import time

import zmq
import numpy as np
import ezmsg.core as ez
from ezmsg.util.debuglog import DebugLog
from ezmsg.util.messages.axisarray import AxisArray
from ezmsg.util.messages.chunker import ArrayChunker
from ezmsg.util.terminate import TerminateOnTotal, TerminateOnTimeout
from ezmsg.util.messagecodec import MessageDecoder

from ezmsg.zmq.pubsub import ZMQPollerUnit, ZMQSenderUnit
from ezmsg.zmq.util import SerializeMessage


def test_poller():
    port = 5557
    topic = "test_poller"
    state = {"running": True, "count": 0}

    comps = {
        "POLLER": ZMQPollerUnit(
            read_addr="tcp://127.0.0.1:" + str(port), zmq_topic=topic
        ),
        "LOGGER": DebugLog(),
        "TERM": TerminateOnTotal(total=10),
    }
    conns = (
        (comps["POLLER"].OUTPUT, comps["LOGGER"].INPUT),
        (comps["LOGGER"].OUTPUT, comps["TERM"].INPUT_MESSAGE),
    )

    def pub_thread():
        ctx = zmq.Context()
        sock = ctx.socket(zmq.PUB)
        sock.bind("tcp://*:" + str(port))
        while state["running"]:
            sock.send(
                b"".join((bytes(topic, "UTF-8"), bytes(json.dumps(state), "UTF-8"))),
                flags=zmq.NOBLOCK,
            )
            state["count"] += 1
            time.sleep(0.05)
        sock.close()
        ctx.term()

    _thread = threading.Thread(target=pub_thread)
    _thread.daemon = True
    _thread.start()

    ez.run(components=comps, connections=conns)

    state["running"] = False
    _thread.join()


def test_sender():
    port = 5555
    topic = "test_sender"
    state = {"running": True, "count": 0}
    ntimes = 100
    nch = 8
    nfeats = 3
    data = np.arange(np.prod((ntimes, nch, nfeats))).reshape((ntimes, nch, nfeats))

    comps = {
        "SRC": ArrayChunker(data=data, chunk_len=10, axis=0, fs=100.0),
        "ENCODE": SerializeMessage(),  # From Any to ZMQMessage
        "PUB": ZMQSenderUnit(
            write_addr="tcp://*:" + str(port), zmq_topic=topic, wait_for_sub=True
        ),
        "TERM": TerminateOnTimeout(time=1.0),
    }
    conns = (
        (comps["SRC"].OUTPUT_SIGNAL, comps["ENCODE"].INPUT),
        (comps["ENCODE"].OUTPUT, comps["PUB"].INPUT),
        (comps["SRC"].OUTPUT_SIGNAL, comps["TERM"].INPUT),
    )

    def sub_thread(res):
        ctx = zmq.Context()
        sock = ctx.socket(zmq.SUB)
        sock.connect("tcp://127.0.0.1:" + str(port))
        # sock.setsockopt(zmq.LINGER, 0)
        sock.setsockopt(zmq.RCVTIMEO, 500)
        sock.setsockopt(zmq.SUBSCRIBE, bytes(topic, "UTF-8"))
        while state["running"]:
            try:
                msg = sock.recv()
            except zmq.Again:
                continue
            payload = msg[len(topic) :].decode("utf-8")
            axarr = json.loads(payload, cls=MessageDecoder)["obj"]
            res.append(axarr)
        sock.close()

    result = []
    _thread = threading.Thread(target=sub_thread, args=(result,))
    _thread.daemon = True
    _thread.start()
    time.sleep(0.1)

    ez.run(components=comps, connections=conns)

    state["running"] = False
    _thread.join()

    assert len(result) == 10
    cat = AxisArray.concatenate(*result, dim="time")
    assert np.array_equal(cat.data, data)



from dataclasses import dataclass
import json
from pickle import PickleBuffer
import typing

import ezmsg.core as ez
from ezmsg.util.messagelogger import log_object
from ezmsg.util.messagecodec import MessageDecoder


class ZeroCopyBytes(bytes):
    def __reduce_ex__(self, protocol):
        if protocol >= 5:
            return type(self)._reconstruct, (PickleBuffer(self),), None
        else:
            # PickleBuffer is forbidden with pickle protocols <= 4.
            return type(self)._reconstruct, (bytes(self),)

    @classmethod
    def _reconstruct(cls, obj):
        with memoryview(obj) as m:
            # Get a handle over the original buffer object
            obj = m.obj
            if isinstance(obj, cls):
                # Original buffer object is a ZeroCopyBytes, return it
                # as-is.
                return obj
            else:
                return cls(obj)


@dataclass
class ZMQMessage:
    data: bytes


def serialize_msg(msg: typing.Any) -> bytes:
    return log_object(msg).encode("utf-8")


"""
The following alternative to serialize_msg might be faster because it doesn't convert numpy arrays to ascii.

class NumpyArrayEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, "tolist"):
            # Likely numpy array to list
            return obj.tolist()
        else:
            return json.JSONEncoder.default(self, obj)


def serialize_msg(msg: typing.Any) -> bytes:
    return json.dumps(asdict(msg), cls=NumpyArrayEncoder).encode("utf-8")
"""


class SerializeMessageSettings(ez.Settings):
    fun: typing.Callable = serialize_msg
    """
    Function to serialize the message. Must take a single argument and return a bytes object.
    """


class SerializeMessage(ez.Unit):
    SETTINGS: SerializeMessageSettings

    INPUT = ez.InputStream(typing.Any)
    OUTPUT = ez.OutputStream(ZMQMessage)

    @ez.subscriber(INPUT)
    @ez.publisher(OUTPUT)
    async def on_message(self, message: typing.Any) -> typing.AsyncGenerator:
        encoded = self.SETTINGS.fun(message)
        yield self.OUTPUT, ZMQMessage(data=encoded)


class DeserializeBytes(ez.Unit):
    INPUT = ez.InputStream(bytes)
    OUTPUT_SIGNAL = ez.OutputStream(typing.Any)

    @ez.subscriber(INPUT)
    @ez.publisher(OUTPUT_SIGNAL)
    async def deserialize(self, msg: ZMQMessage) -> typing.AsyncGenerator:
        decoded_msg = json.loads(msg.data, cls=MessageDecoder)["obj"]
        yield self.OUTPUT_SIGNAL, decoded_msg



import json
from pathlib import Path
import tempfile

import numpy as np
import ezmsg.core as ez
from ezmsg.util.messages.axisarray import AxisArray
from ezmsg.util.messagelogger import log_object, MessageLogger
from ezmsg.util.messagecodec import MessageDecoder, message_log
from ezmsg.util.messages.chunker import ArrayChunker
from ezmsg.util.terminate import TerminateOnTotal

from ezmsg.zmq.util import SerializeMessage, DeserializeBytes


def test_encdec():
    msg = AxisArray(
        data=np.arange(np.prod((3, 4, 5))).reshape(3, 4, 5),
        dims=["time", "ch", "feat"],
        axes={
            "time": AxisArray.TimeAxis(fs=1000.0, offset=0.0),
            "ch": AxisArray.CoordinateAxis(np.array(["a", "b", "c", "d"]), dims=["ch"]),
            "feat": AxisArray.CoordinateAxis(
                data=np.array(["f1", "f2", "f3", "f4", "f5"]), dims=["feat"]
            ),
        },
        key="test_log_object",
    )
    encoded_msg = log_object(msg)
    decoded_msg = json.loads(encoded_msg, cls=MessageDecoder)["obj"]
    assert np.array_equal(decoded_msg.data, msg.data)
    assert np.array_equal(decoded_msg.axes["ch"].data, msg.axes["ch"].data)
    assert np.array_equal(decoded_msg.axes["feat"].data, msg.axes["feat"].data)
    assert decoded_msg.key == msg.key


def test_encdec_system():
    ntimes = 100
    nch = 8
    nfeats = 3
    data = np.arange(np.prod((ntimes, nch, nfeats))).reshape((ntimes, nch, nfeats))
    file_path = Path(tempfile.gettempdir())
    file_path = file_path / Path("test_serialize_message.txt")

    comps = {
        "SRC": ArrayChunker(data=data, chunk_len=10, axis=0, fs=100.0),
        "SERIALIZE": SerializeMessage(),
        "DESERIALIZE": DeserializeBytes(),
        "LOG": MessageLogger(output=file_path),
        "TERM": TerminateOnTotal(total=10),
    }
    conns = (
        (comps["SRC"].OUTPUT_SIGNAL, comps["SERIALIZE"].INPUT),
        (comps["SERIALIZE"].OUTPUT, comps["DESERIALIZE"].INPUT),
        (comps["DESERIALIZE"].OUTPUT_SIGNAL, comps["LOG"].INPUT_MESSAGE),
        (comps["LOG"].OUTPUT_MESSAGE, comps["TERM"].INPUT_MESSAGE),
    )
    ez.run(components=comps, connections=conns)

    result = list(message_log(file_path))
    assert len(result) == 10
    cat = AxisArray.concatenate(*result, dim="time")
    assert np.array_equal(cat.data, data)

    file_path.unlink(missing_ok=True)



import asyncio
import typing

import zmq
import zmq.asyncio
import ezmsg.core as ez
from zmq.utils.monitor import parse_monitor_message

from .util import ZMQMessage


class ZMQRepSettings(ez.Settings):
    addr: str


class ZMQRepState(ez.State):
    context: zmq.asyncio.Context
    socket: zmq.asyncio.Socket
    queue: asyncio.Queue


class ZMQRep(ez.Unit):
    OUTPUT = ez.OutputStream(ZMQMessage)

    SETTINGS = ZMQRepSettings
    STATE = ZMQRepState

    def initialize(self) -> None:
        self.STATE.context = zmq.asyncio.Context()
        self.STATE.socket = self.STATE.context.socket(zmq.REP)
        ez.logger.debug(f"{self}:binding to {self.SETTINGS.addr}")
        self.STATE.socket.bind(self.SETTINGS.addr)
        self.STATE.queue = asyncio.Queue()

    def shutdown(self) -> None:
        self.STATE.socket.close()
        self.STATE.context.term()

    def _handle_req(self, data: bytes) -> bytes:
        return data

    @ez.task
    async def zmq_rep(self) -> None:
        while True:
            data = await self.STATE.socket.recv()
            response = self._handle_req(data)
            await self.STATE.socket.send(response)
            self.STATE.queue.put_nowait(data)

    @ez.publisher(OUTPUT)
    async def send_reqs(self) -> typing.AsyncGenerator:
        while True:
            data = await self.STATE.queue.get()
            yield self.OUTPUT, ZMQMessage(data)


class ZMQReqSettings(ez.Settings):
    addr: str


class ZMQReqState(ez.State):
    context: zmq.asyncio.Context
    socket: zmq.asyncio.Socket
    monitor: zmq.asyncio.Socket


class ZMQReq(ez.Unit):
    INPUT = ez.InputStream(ZMQMessage)
    OUTPUT = ez.OutputStream(ZMQMessage)

    SETTINGS = ZMQReqSettings
    STATE = ZMQReqState

    def initialize(self) -> None:
        self.STATE.context = zmq.asyncio.Context()
        self.STATE.socket = self.STATE.context.socket(zmq.REQ)
        self.STATE.monitor = self.STATE.socket.get_monitor_socket()
        ez.logger.debug(f"{self}:connecting to {self.SETTINGS.addr}")
        self.STATE.socket.connect(self.SETTINGS.addr)
        self._has_server = False

    def shutdown(self) -> None:
        self.STATE.monitor.close()
        self.STATE.socket.close()
        self.STATE.context.term()

    @ez.task
    async def _socket_monitor(self) -> None:
        while True:
            monitor_result = await self.STATE.monitor.poll(100, zmq.POLLIN)
            if monitor_result:
                data = await self.STATE.monitor.recv_multipart()
                evt = parse_monitor_message(data)
                event = evt["event"]
                if event == zmq.EVENT_CONNECTED:
                    self._has_server = True
                elif event == zmq.EVENT_DISCONNECTED:
                    self._has_server = False

    @ez.subscriber(INPUT, zero_copy=True)
    @ez.publisher(OUTPUT)
    async def send_req(self, msg: ZMQMessage) -> None:
        if self._has_server:
            await self.STATE.socket.send(msg.data)
            response = await self.STATE.socket.recv()
            yield self.OUTPUT, ZMQMessage(response)



import asyncio
from pathlib import Path
import tempfile
import threading

import zmq
import ezmsg.core as ez
from ezmsg.util.messagelogger import MessageLogger
from ezmsg.util.messagecodec import message_log
from ezmsg.util.terminate import TerminateOnTotal
from zmq.utils.monitor import parse_monitor_message

from ezmsg.zmq.repreq import ZMQRep, ZMQReq
from ezmsg.zmq.util import ZMQMessage


def test_rep():
    port = 5557
    state = {"running": True, "count": 0, "connected": False}
    file_path = Path(tempfile.gettempdir())
    file_path = file_path / Path("test_rep.txt")

    comps = {
        "ZREP": ZMQRep(addr="tcp://*:" + str(port)),
        "LOGGER": MessageLogger(output=file_path),
        "TERM": TerminateOnTotal(total=10),
    }
    conns = (
        (comps["ZREP"].OUTPUT, comps["LOGGER"].INPUT_MESSAGE),
        (comps["LOGGER"].OUTPUT_MESSAGE, comps["TERM"].INPUT_MESSAGE),
    )

    def req_thread():
        ctx = zmq.Context()
        sock = ctx.socket(zmq.REQ)
        sock.setsockopt(zmq.SNDTIMEO, 500)
        sock.setsockopt(zmq.RCVTIMEO, 500)
        sock.connect("tcp://localhost:" + str(port))

        monitor = sock.get_monitor_socket()
        # It would be better to wait for a connection.
        while state["running"]:
            if not state["connected"]:
                monitor_result = monitor.poll(100)
                if monitor_result:
                    data = monitor.recv_multipart()
                    evt = parse_monitor_message(data)
                    if evt["event"] == zmq.EVENT_CONNECTED:
                        state["connected"] = True
            if state["connected"]:
                try:
                    sock.send(b"Hello")
                    msg = sock.recv()
                    assert msg == b"Hello"
                except zmq.error.Again:
                    # Remote has disconnected. Try again.
                    # If this happened during `send`, then we
                    #  might be in a bad state because we can't
                    #  send twice without a successful recv in between.
                    continue
        monitor.close()
        sock.close()
        ctx.term()

    _thread = threading.Thread(target=req_thread)
    _thread.daemon = True
    _thread.start()

    ez.run(components=comps, connections=conns)

    state["running"] = False
    _thread.join()

    result = list(message_log(file_path))
    for msg in result:
        assert msg.data == b"Hello"

    file_path.unlink(missing_ok=True)


def test_req():
    port = 5557
    state = {"running": True, "count": 0}
    file_path = Path(tempfile.gettempdir())
    file_path = file_path / Path("test_req.txt")

    class DummyReqSource(ez.Unit):
        OUTPUT = ez.OutputStream(ZMQMessage)

        @ez.publisher(OUTPUT)
        async def send_reqs(self):
            while True:
                yield self.OUTPUT, ZMQMessage(b"Hello")
                await asyncio.sleep(0.1)

    comps = {
        "SRC": DummyReqSource(),
        "ZREQ": ZMQReq(addr="tcp://localhost:" + str(port)),
        "LOGGER": MessageLogger(output=file_path),
        "TERM": TerminateOnTotal(total=10),
    }
    conns = (
        (comps["SRC"].OUTPUT, comps["ZREQ"].INPUT),
        (comps["ZREQ"].OUTPUT, comps["LOGGER"].INPUT_MESSAGE),
        (comps["LOGGER"].OUTPUT_MESSAGE, comps["TERM"].INPUT_MESSAGE),
    )

    def rep_thread():
        ctx = zmq.Context()
        sock = ctx.socket(zmq.REP)
        sock.setsockopt(zmq.RCVTIMEO, 500)
        sock.setsockopt(zmq.SNDTIMEO, 500)
        sock.bind("tcp://*:" + str(port))
        while state["running"]:
            try:
                msg = sock.recv()
                assert msg == b"Hello"
                sock.send(msg)
            except zmq.error.Again:
                # Remote has disconnected. Try again.
                continue
        sock.close()
        ctx.term()

    _thread = threading.Thread(target=rep_thread)
    _thread.daemon = True
    _thread.start()

    ez.run(components=comps, connections=conns)

    state["running"] = False
    _thread.join()

    for msg in message_log(file_path):
        assert msg.data == b"Hello"

    file_path.unlink(missing_ok=True)
